from django.contrib import admin
from allauth_sniperlinks.models import SniperLink

@admin.register(SniperLink)
class SniperLinkAdmin(admin.ModelAdmin):
    list_display = ['email_object', 'mx_domain', 'mail_provider']



import time
import re
import tld
import dns.resolver
from django.contrib.staticfiles import finders
from django.templatetags.static import static
from django.core.cache import cache
from .models import MailProviders

INVALID_EMAIL_FLAG = 'INVALIDEMAIL'
USER_SNIPERLINK_CACHE_PREFIX = 'sniperlinks_user_'
USER_SNIPERLINK_CACHE_KEY = USER_SNIPERLINK_CACHE_PREFIX + '{}'
SNIPERLINKS_CACHE_VERSION_KEY = 'sniperlinks_version'


def get_sniperlink_cache_version():
    return cache.get(SNIPERLINKS_CACHE_VERSION_KEY)

def bump_sniperlink_cache_version():
    ts_val = int(time.time())
    # attempt to cache forever (but may get bumped), set to current timestamp
    cache.set(SNIPERLINKS_CACHE_VERSION_KEY, ts_val, timeout=None)
    return ts_val

def get_sniper_link_img(mail_provider):
    try:
        filename = 'allauth_sniperlinks/provider_icons/{}.svg'.format(
            dict(MailProviders.choices)[mail_provider]
        )
        if finders.find(filename):
            filepath = static(filename)
        else: 
            filepath = static('allauth_sniperlinks/provider_icons/generic.svg')
    except KeyError:
        filepath = static('allauth_sniperlinks/provider_icons/generic.svg')
        
    return filepath


def get_mx_domian_and_provider(email_address):
    if email_address:
        match = re.match(r'(.*)@(.*)', email_address)
        if match is None or len(match.groups()) != 2:
            return INVALID_EMAIL_FLAG, None
        domain = match.group(2).lower()
        try:
            mx_records = dns.resolver.resolve(domain, 'MX')
            mx_record = mx_records[0]
            for r in mx_records:
                if r.preference < mx_record.preference:
                    mx_record = r
        except Exception as e:
            return INVALID_EMAIL_FLAG, None
        try:
            mx_res_obj = tld.get_tld(
                mx_record.exchange.to_text(),
                fix_protocol=True,
                as_object=True,
            )
            email_res_obj = tld.get_tld(
                domain,
                fix_protocol=True,
                as_object=True
            )
            mx_domain = mx_res_obj.fld.lower()
            mail_provider = get_mail_provider(email_res_obj, mx_res_obj)
            return mx_domain, mail_provider
        except tld.exceptions.TldDomainNotFound as e:
            return INVALID_EMAIL_FLAG, None
        except Exception as e:
            return INVALID_EMAIL_FLAG, None


# TODO - changte these to regexes in case MX records span multiple domains or change
# TODO - add nightly tests to check for changes in MX domains
def get_mail_provider(email_res_obj, mx_res_obj):
    mx_domain = mx_res_obj.domain.lower()
    email_domain = email_res_obj.domain.lower()
    
    if mx_domain == 'google' or mx_domain == 'googlemail':
        if email_domain == 'gmail':
            return MailProviders.GMAIL
        else:
            return MailProviders.GSUITE
    
    if mx_domain == 'yahoodns':
        if email_domain == 'aol' or mx_domain == 'cs':
            return MailProviders.AOL
        if email_domain == 'yahoo':
            return MailProviders.YAHOO
        if email_domain == 'rogers':
            return MailProviders.ROGERS
        if email_domain == 'frontier':
            return MailProviders.FRONTIER
        else:
            return MailProviders.YAHOO

    if mx_domain == 'outlook':
        return MailProviders.OUTLOOK

    if mx_domain == 'icloud':
        return MailProviders.ICLOUD
    
    if mx_domain == 'charter':
        return MailProviders.CHARTER
    
    if mx_domain == 'centurylink':
        return MailProviders.CENTURYLINK
    
    if mx_domain == '1and1':
        return MailProviders.ONEANDONE
    
    if mx_domain == 'cloudfilter':
        if email_domain == 'shaw':
            return MailProviders.SHAW
        if email_domain == 'cox':
            return MailProviders.COX
    
    if mx_domain == 'comcast':
        return MailProviders.COMCAST

    if mx_domain == 'messagingengine':
        return MailProviders.FASTMAIL
    
    if mx_domain == 'frontiernet':
        return MailProviders.FRONTIER
    
    if mx_domain == 'mail':
        return MailProviders.MAILDOTCOM

    if mx_domain == 'oxsus-vadesecure':
        return MailProviders.EARTHLINK
    
    if mx_domain == 'prodigy':
        return MailProviders.PRODIGY
    
    if mx_domain == 'protonmail':
        return MailProviders.PROTONMAIL

    if mx_domain == 'untd':
        return MailProviders.JUNO
    
    if mx_domain == 'netaddress':
        return MailProviders.NETADDRESS
    
    if mx_domain == 'videotron':
        return MailProviders.VIDEOTRON
    
    if mx_domain == 'windstream':
        return MailProviders.WINDSTREAM
    
    if mx_domain == 'zoho':
        return MailProviders.ZOHO

    return MailProviders.UNKNOWN


# TODO - update these where possible to have better mobile links to email apps
def get_sniper_link(mail_provider, email_address, outgoing_email_address):
    if mail_provider == MailProviders.UNKNOWN or outgoing_email_address is None:
        return ""

    match = re.match(r'(.*)@(.*)', outgoing_email_address)
    if match is None or len(match.groups()) != 2:
            return ""
    sending_domain = match.group(2).lower()

    match = re.match(r'(.*)@(.*)', email_address)
    if match is None or len(match.groups()) != 2:
        receiving_domain = ""
    else:
        receiving_domain = match.group(2).lower()

    if mail_provider == MailProviders.GMAIL:
        return "https://mail.google.com/mail/u/?#search/from%3A%40{}+in%3Aanywhere".format(
            sending_domain
        )
    
    if mail_provider == MailProviders.GSUITE:
        if receiving_domain:
            return "https://mail.google.com/a/{}/#search/from%3A%40{}+in%3Aanywhere".format(
                receiving_domain,
                sending_domain
            )
        else:
            return "https://mail.google.com/mail/u/?#search/from%3A%40{}+in%3Aanywhere".format(
                sending_domain
            )

    if mail_provider == MailProviders.YAHOO:
        return "https://mail.yahoo.com/d/search/keyword=from%3A{}".format(
            outgoing_email_address
        )
    
    if mail_provider == MailProviders.ROGERS:
        # has search but routes through Rogers-specific login if needed.
        # Seems search param will not be maintained if login required, but should work if already logged in.
        return "https://mail.yahoo.com/d/search/keyword=from%3A{}?.intl=ca&.partner=rogers-acs&.lang=en-CA".format(
            outgoing_email_address
        )

    if mail_provider == MailProviders.FRONTIER:
        # has search but routes through Frontier-specific login if needed.
        # (Unclear if search param will be maintained if login required, but should work if already logged in.)
        return "https://mail.yahoo.com/d/search/keyword=from%3A{}?.intl=us&.partner=ftr&.lang=en-US".format(
            outgoing_email_address
        )
    
    if mail_provider == MailProviders.PRODIGY:
        # has search but routes through ATT-specific login if needed.
        # (Unclear if search param will be maintained if login required, but should work if already logged in.)
        return "https://mail.yahoo.com/d/search/keyword=from%3A{}?.intl=us&.partner=sbc&.lang=en-US".format(
            outgoing_email_address
        )

    if mail_provider == MailProviders.AOL:
        return "https://https://mail.aol.com/"

    if mail_provider == MailProviders.OUTLOOK:
        return "https://outlook.live.com/"

    # really need to do this one based on device, open app on mobile
    if mail_provider == MailProviders.ICLOUD:
        return "https://www.icloud.com/mail"
    
    if mail_provider == MailProviders.CHARTER:
        return "http://webmail.spectrum.net/"
    
    if mail_provider == MailProviders.CENTURYLINK:
        # redirect to login only works for recent timestamps, otherwise goes to
        # main page
        ts = int(time.time())
        return "http://centurylink.net/zmail/index.php?autologin=true&ts={}".format(ts)

    if mail_provider == MailProviders.ONEANDONE:
        return "https://mail.ionos.ca/"

    if mail_provider == MailProviders.SHAW:
        return "https://webmail.shaw.ca/"

    if mail_provider == MailProviders.COX:
            return "https://myemail.cox.net/"
    
    if mail_provider == MailProviders.COMCAST:
        return "https://xfinityconnect.email.comcast.net/" 

    if mail_provider == MailProviders.FASTMAIL:
        # Supports searching all folders with in:*
        return "https://www.fastmail.com/mail/search:in%3A*+from%3A{}/".format(
            sending_domain
        )

    if mail_provider == MailProviders.MAILDOTCOM:
        # MAIL.COM annoyingly doesn't let you use a generic link, so it will redirect to mail.com 
        # and make people log in every time. Stupid.
        return "https://navigator-lxa.mail.com/"

    if mail_provider == MailProviders.PROTONMAIL:
        # Supports searching all with all-mail
        return "https://mail.protonmail.com/u/0/all-mail#from={}".format(
            sending_domain
        )

    if mail_provider == MailProviders.JUNO:
        return "https://webmaila.juno.com/webmail/new/"

    # if mail_provider == MailProviders.EARTHLINK:
    # if mail_provider == MailProviders.NETADDRESS:
    # if mail_provider == MailProviders.VIDEOTRON:
    # if mail_provider == MailProviders.WINDSTREAM:
    # if mail_provider == MailProviders.ZOHO:

    return ""


from django import template
from django.utils.safestring import mark_safe
from django.template.loader import render_to_string


register = template.Library()


@register.simple_tag
def get_sniperlink_for_email(sniperlinks, email):
    if sniperlinks and email in sniperlinks:
        return sniperlinks[email]


@register.simple_tag(takes_context=True)
def unverified_email_banner(context, banner_class="", link_class=""):
    print("Hi mom!")
    sniperlinks = context['sniperlinks']
    context['banner_class'] = banner_class
    context['link_class'] = link_class
    html = ''
    # Check if the user has any unverified email addresses from allauth
    if sniperlinks:
        # Loop through the sniperlinks in the context.
        for email in sniperlinks:
            banner_context = context.flatten()
            link = sniperlinks[email]['link']
            if link:
                banner_context['email'] = email
                banner_context['link'] = link
                banner_context['img'] = sniperlinks[email]['img']
                html += render_to_string('allauth_sniperlinks/banner.html', banner_context)

    # Return the HTML as a safe string
    return mark_safe(html)



from django.conf import settings
from django.apps import apps as django_apps
from django.core.exceptions import ImproperlyConfigured

VERIFICATION_SENDER = getattr(settings, "ALLAUTH_SNIPERLINKS_VERIFICATION_SENDER", settings.DEFAULT_FROM_EMAIL)
BANNER_ONLY_PRIMARY = getattr(settings, "ALLAUTH_SNIPERLINKS_BANNER_ONLY_PRIMARY", True)
SAFE_TEMPLATES = getattr(
    settings, "ALLAUTH_SNIPERLINKS_SAFE_TEMPLATES",
    [
        "account/messages/" "email_confirmation_sent.txt",
    ]
)


# Generated by Django 3.2.10 on 2022-12-02 20:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('allauth_sniperlinks', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sniperlink',
            name='mail_provider',
            field=models.CharField(blank=True, choices=[('UNKNOWN', 'Unknown'), ('GMAIL', 'Gmail'), ('GSUITE', 'Gsuite'), ('YAHOO', 'Yahoo'), ('AOL', 'AOL'), ('ROGERS', 'ROGERS'), ('OUTLOOK', 'Outlook'), ('ICLOUD', 'iCloud'), ('CHARTER', 'Charter'), ('CENTURYLINK', 'CenturyLink'), ('ONEANDONE', '1and1'), ('COMCAST', 'Comcast'), ('COX', 'Cox'), ('SHAW', 'Shaw'), ('FASTMAIL', 'Fastmail'), ('FRONTIER', 'Frontier'), ('MAILDOTCOM', 'Mail_com'), ('EARTHLINK', 'Earthlink'), ('PRODIGY', 'Prodigy'), ('PROTONMAIL', 'ProtonMail'), ('JUNO', 'Juno'), ('NETADDRESS', 'NetAddress'), ('VIDEOTRON', 'Videotron'), ('WINDSTREAM', 'Windstream'), ('ZOHO', 'Zoho')], default='UNKNOWN', max_length=30, null=True),
        ),
    ]



from django.core.cache import cache
from django.db.models.signals import post_save, pre_save
from django.dispatch import receiver
from allauth.account.models import EmailAddress
from allauth_sniperlinks import app_settings
from allauth_sniperlinks.utils import get_mx_domian_and_provider, get_sniperlink_cache_version
from allauth_sniperlinks.utils import USER_SNIPERLINK_CACHE_KEY
from allauth_sniperlinks.models import SniperLink


@receiver(pre_save, sender=EmailAddress)
def cache_previous_email(sender, instance, *args, **kwargs):
    original_email = None
    if instance.id:
        original_email = EmailAddress.objects.get(pk=instance.id).email
    
    instance.__original_email = original_email


@receiver(post_save, sender=EmailAddress)
def update_sniperlink_and_clear_cache(sender, instance, created, **kwargs):
    # update the sniper link or create it if it doesn't exist
    if not hasattr(instance, 'sniperlink'):
        mx_domain, mail_provider = get_mx_domian_and_provider(instance.email)
        SniperLink.objects.create(
            email_object=instance,
            mx_domain=mx_domain,
            mail_provider=mail_provider
        )
    elif instance.__original_email != instance.email:
        mx_domain, mail_provider = get_mx_domian_and_provider(instance.email)
        instance.sniperlink.mx_domain = mx_domain
        instance.sniperlink.mail_provider = mail_provider
        instance.sniperlink.save()

    # invalidate the cached sniperlinks for the user
    # these are re-generated by context processor on next request
    version = get_sniperlink_cache_version()
    cache.delete(USER_SNIPERLINK_CACHE_KEY.format(instance.user.id), version=version)


from django.db import models
from django.utils.translation import gettext_lazy as _
from allauth.account.models import EmailAddress

# Create your models here.
class MailProviders(models.TextChoices):
    UNKNOWN = 'UNKNOWN', _('Unknown')
    GMAIL = 'GMAIL', _('Gmail')
    GSUITE = 'GSUITE', _('Gsuite')
    YAHOO = 'YAHOO', _('Yahoo')
    AOL = 'AOL', _('AOL')
    ROGERS = 'ROGERS', _('ROGERS')
    OUTLOOK = 'OUTLOOK', _('Outlook')
    ICLOUD = 'ICLOUD', _('iCloud')
    CHARTER = 'CHARTER', _('Charter')
    CENTURYLINK = 'CENTURYLINK', _('CenturyLink')
    ONEANDONE = 'ONEANDONE', _('1and1')
    COMCAST = 'COMCAST', _('Comcast')
    COX = 'COX', _('Cox')
    SHAW = 'SHAW', _('Shaw')
    FASTMAIL = 'FASTMAIL', _('Fastmail')
    FRONTIER = 'FRONTIER', _('Frontier')
    MAILDOTCOM = 'MAILDOTCOM', _('Mail_com')
    EARTHLINK = 'EARTHLINK', _('Earthlink')
    PRODIGY = 'PRODIGY', _('Prodigy')
    PROTONMAIL = 'PROTONMAIL', _('ProtonMail')
    JUNO = 'JUNO', _('Juno')
    NETADDRESS = 'NETADDRESS', _('NetAddress')  # USA.NET
    VIDEOTRON = 'VIDEOTRON', _('Videotron')
    WINDSTREAM = 'WINDSTREAM', _('Windstream')  # (Seems same as Century Link)
    ZOHO = 'ZOHO', _('Zoho')


class SniperLink(models.Model):
    email_object = models.OneToOneField(
        EmailAddress,
        on_delete=models.CASCADE,
    )
    mx_domain = models.CharField(
        max_length=100,
        null=True,
        blank=True
    )
    mail_provider = models.CharField(
        max_length=30,
        null=True,
        blank=True,
        choices=MailProviders.choices,
        default=MailProviders.UNKNOWN,
    )



from django.core.management.base import BaseCommand
from django.core.cache import cache

from allauth_sniperlinks.utils import bump_sniperlink_cache_version

class Command(BaseCommand):
    def handle(self, *args, **kwargs):
        bump_sniperlink_cache_version()
        self.stdout.write('Sniperlink cache version incremented\n')


from django.conf import settings
from django.contrib import messages
from django.template.loader import render_to_string
from django.template import TemplateDoesNotExist
from django.utils.safestring import mark_safe
from allauth.account.adapter import DefaultAccountAdapter
from allauth_sniperlinks import app_settings


class SniperLinkAccountAdapter(DefaultAccountAdapter):
    def add_message(
        self,
        request,
        level,
        message_template,
        message_context=None,
        extra_tags="",
    ):
        """
        Wrapper of `django.contrib.messages.add_message`, that reads
        the message text from a template.
        """
        if "django.contrib.messages" in settings.INSTALLED_APPS:
            try:
                if message_context is None:
                    message_context = {}
                message = render_to_string(
                    message_template,
                    message_context,
                    self.request,
                ).strip()
                if message:
                    if message_template in app_settings.SAFE_TEMPLATES:
                        message = mark_safe(message)
                    messages.add_message(request, level, message, extra_tags=extra_tags)
            except TemplateDoesNotExist:
                pass


# Generated by Django 3.2.10 on 2022-12-01 20:37

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('account', '0002_email_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='SniperLink',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('mx_domain', models.CharField(blank=True, max_length=100, null=True)),
                ('mail_provider', models.CharField(choices=[('UNKNOWN', 'Unknown'), ('GMAIL', 'Gmail'), ('GSUITE', 'Gsuite'), ('YAHOO', 'Yahoo'), ('AOL', 'AOL'), ('ROGERS', 'ROGERS'), ('OUTLOOK', 'Outlook'), ('ICLOUD', 'iCloud'), ('CHARTER', 'Charter'), ('CENTURYLINK', 'CenturyLink'), ('ONEANDONE', '1and1'), ('COMCAST', 'Comcast'), ('COX', 'Cox'), ('SHAW', 'Shaw'), ('FASTMAIL', 'Fastmail'), ('FRONTIER', 'Frontier'), ('MAILDOTCOM', 'Mail_com'), ('EARTHLINK', 'Earthlink'), ('PRODIGY', 'Prodigy'), ('PROTONMAIL', 'ProtonMail'), ('JUNO', 'Juno'), ('NETADDRESS', 'NetAddress'), ('VIDEOTRON', 'Videotron'), ('WINDSTREAM', 'Windstream'), ('ZOHO', 'Zoho')], default='UNKNOWN', max_length=30)),
                ('email_object', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='account.emailaddress')),
            ],
        ),
    ]



from django.core.cache import cache
from allauth_sniperlinks.models import SniperLink
from allauth_sniperlinks.utils import get_sniper_link, get_sniper_link_img, bump_sniperlink_cache_version
from allauth_sniperlinks.utils import USER_SNIPERLINK_CACHE_KEY, get_sniperlink_cache_version
from allauth_sniperlinks.app_settings import VERIFICATION_SENDER, BANNER_ONLY_PRIMARY


def unverified_email_sniperlinks(request):
    """
    Adds sniperlinks for any unverified emails for the currently logged in user
    to context.
    """
    if request.user.is_authenticated:
        sniperlinks_version = get_sniperlink_cache_version()
        if sniperlinks_version is None:
            sniperlinks_version = bump_sniperlink_cache_version()
        sniperlinks_context = cache.get(USER_SNIPERLINK_CACHE_KEY.format(request.user.id), version=sniperlinks_version)
        if sniperlinks_context is None:
            sniperlinks_context = {}
            unverified_email_sls = SniperLink.objects.filter(
                email_object__user=request.user, 
                email_object__verified=False
            ).order_by('email_object__primary')
            if BANNER_ONLY_PRIMARY:
                unverified_email_sls = unverified_email_sls.filter(email_object__primary=True)
            for sl in unverified_email_sls:
                sniperlinks_context[sl.email_object.email] = {
                    'link': get_sniper_link(sl.mail_provider, sl.email_object.email, VERIFICATION_SENDER),
                    'img': get_sniper_link_img(sl.mail_provider)
                }
            # Cache entry is invalidated via signal handler when any of the user's email entries change
            cache.set(
                USER_SNIPERLINK_CACHE_KEY.format(request.user.id),
                sniperlinks_context,
                version=sniperlinks_version
            )
        return {
            'sniperlinks': sniperlinks_context
        }
    else:
        return {}